/**
 * gclc-socket:net.bigeon.gclc.socket.PlugableConsoleInput.java
 * Created on: Nov 18, 2017
 */
package net.bigeon.gclc.socket;

/*-
 * #%L
 * GCLC Socket
 * %%
 * Copyright (C) 2016 - 2018 Bigeon
 * %%
 * This software is governed by the CeCILL  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 * #L%
 */
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.nio.charset.StandardCharsets;
import java.util.function.Supplier;
import java.util.logging.Level;
import java.util.logging.Logger;

import net.bigeon.gclc.manager.ConsoleInput;
import net.bigeon.gclc.tools.ConstantString;
import net.bigeon.gclc.utils.ReadingRunnable;

/** A console input where the stream can be plugged.
 * <p>
 * This pluggable console input accepts an input and output to be connected to
 * it. The connection cannot be concurrent, which mean that any connected stream
 * must be disconnected before a new call to
 * {@link #connect(InputStream, PrintStream)} is done.
 * <p>
 * On connection, if the prompt is in wait, this console input will provide the
 * prompting message.
 *
 * @author Emmanuel Bigeon */
public final class PluggableConsoleInput implements ConsoleInput {
    /** The ten constant. */
    private static final int    TENTH       = 10;
    /** Class logger. */
    private static final Logger LOGGER      = Logger
            .getLogger(PluggableConsoleInput.class.getName());
    /** The default time out. */
    private static final long   TIMEOUT     = 100;
    /** The prompting. */
    private boolean             prompting   = false;
    /** If the element is closed. */
    private boolean             closed      = false;
    /** The default prompt. */
    private Supplier<String>    prompt      = new ConstantString("> "); //$NON-NLS-1$
    /** If the input is plugged or buffering. */
    private boolean             connected   = false;
    /** The current connexion (if any). */
    private ReadingRunnable     connexion;
    /** The interrupted status for prompts. */
    private boolean             interrupted = false;
    /** The last hint hint. */
    private String              hint;
    /** The output for hints. */
    private PrintStream         output;

    // Locks
    /** The lock for connexion and disconnexion of actual streams. */
    private final Object connexionLock = new Object();

    /** Create the pluggable console input. */
    public PluggableConsoleInput() {
        // do nothing
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.manager.ConsoleInput#close() */
    @Override
    public void close() {
        closed = true;
    }

    /** Connect the given input stream to the input and output to the hints writing.
     *
     * @param stream the input stream
     * @param out the output for hints.
     * @throws IOException if the input is already connected. */
    public void connect(final InputStream stream, final PrintStream out)
            throws IOException {
        synchronized (connexionLock) {
            if (connected) {
                // Cannot be connected to several sources
                throw new IOException("Input already connected to an input stream"); //$NON-NLS-1$
            }
            output = out;
            if (prompting) {
                // print the hint, to indicate we are waiting for a user input.
                out.print(hint);
                out.flush();
            }
            final InputStreamReader streamReader = new InputStreamReader(stream,
                    StandardCharsets.UTF_8);
            final BufferedReader reader = new BufferedReader(streamReader);
            connexion = new ReadingRunnable(reader);
            final Thread th = new Thread(connexion, "GCLC Socket - Read input"); //$NON-NLS-1$
            th.start();
            connexionLock.notifyAll();
            connected = true;
        }
    }

    /** Disconnect the current input and hint output. */
    public synchronized void disconnect() {
        synchronized (connexionLock) {
            if (!connected) {
                return;
            }
            connected = false;
            connexion.setRunning(false);
        }
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.manager.ConsoleInput#getPrompt() */
    @Override
    public Supplier<String> getPrompt() {
        return prompt;
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.manager.ConsoleInput#interruptPrompt() */
    @Override
    public void interruptPrompt() {
        interrupted = true;
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.manager.ConsoleInput#isClosed() */
    @Override
    public boolean isClosed() {
        return closed;
    }

    /** Test if a prompt is occuring.
     *
     * @return the prompting */
    public boolean isPrompting() {
        return prompting;
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.manager.ConsoleInput#prompt() */
    @Override
    public String prompt() throws IOException {
        return prompt(prompt.get());
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.manager.ConsoleInput#prompt(long) */
    @Override
    public String prompt(final long timeout) throws IOException {
        return prompt(prompt.get(), timeout);
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.manager.ConsoleInput#prompt(java.lang.String) */
    @Override
    public String prompt(final String message) throws IOException {
        if (closed) {
            throw new IOException();
        }
        preparePrompt(message);
        String res = null;
        while (res == null && !interrupted) {
            try {
                // Wait for a message or a connection with message
                res = waitMessageOrConnexion(TIMEOUT, TIMEOUT / TENTH);
            } catch (final InterruptedException e) {
                LOGGER.log(Level.FINE, "Interruption of thread", e); //$NON-NLS-1$
                Thread.currentThread().interrupt();
            }
            if (closed) {
                throw new IOException();
            }
        }
        prompting = false;
        return res;
    }

    /** Prepare the new prompt.
     * 
     * @param message the message to indicate request of prompt. */
    private void preparePrompt(final String message) {
        prompting = true;
        // hold the message
        synchronized (connexionLock) {
            hint = message;
            if (connected) {
                // print the message
                output.print(message);
                output.flush();
            }
        }
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.manager.ConsoleInput#prompt(java.lang.String, long) */
    @Override
    public String prompt(final String message, final long timeout) throws IOException {
        if (closed) {
            throw new IOException();
        }
        preparePrompt(message);
        String res = null;
        final long tic = System.currentTimeMillis();
        long time = System.currentTimeMillis() - tic;
        while (res == null && !interrupted && time < timeout) {
            try {
                res = waitMessageOrConnexion(timeout - time, (timeout - time) / TENTH);
            } catch (final InterruptedException e) {
                LOGGER.log(Level.FINE, "Interruption of thread", e); //$NON-NLS-1$
                Thread.currentThread().interrupt();
            }
            time = System.currentTimeMillis() - tic;
            if (closed) {
                throw new IOException();
            }
        }
        prompting = false;
        return res;
    }

    /* (non-Javadoc)
     * @see net.bigeon.gclc.manager.ConsoleInput#setPrompt(java.lang.String) */
    @Override
    public void setPrompt(final String prompt) {
        setPrompt(new ConstantString(prompt));
    }

    /* (non-Javadoc)
     * @see net.bigeon.gclc.manager.ConsoleInput#setPrompt(net.bigeon.gclc.tools.
     * StringProvider) */
    @Override
    public void setPrompt(final Supplier<String> prompt) {
        this.prompt = prompt;
    }

    /** Wait for a hint or connection.
     *
     * @param messageTimeout the timeout on the current connection hint waiting
     * @param connexionTimeout the timeout on the new connection wait
     * @return the hint, or null if not connected or timed out.
     * @throws IOException if the reading failed.
     * @throws InterruptedException if the wait was interrupted */
    private String waitMessageOrConnexion(final long messageTimeout,
            final long connexionTimeout) throws IOException, InterruptedException {
        synchronized (connexionLock) {
            if (connected) {
                return connexion.getNextMessage(messageTimeout);
            }
            connexionLock.wait(connexionTimeout);
        }
        return null;
    }
}
